@@ -1,3 +1,5 @@
+from functools import wraps
+
 import torch
 from torch import nn, einsum
 import torch.nn.functional as F
@@ -12,6 +14,17 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
+def cache_fn(f):
+    cache = None
+    @wraps(f)
+    def cached_fn(*args, **kwargs):
+        nonlocal cache
+        if cache is not None:
+            return cache
+        cache = f(*args, **kwargs)
+        return cache
+    return cached_fn
+
 def fourier_encode(x, num_encodings = 4):
     x = x.unsqueeze(-1)
     device, dtype, orig_x = x.device, x.dtype, x
@@ -70,7 +83,7 @@ class Attention(nn.Module):
             nn.Dropout(dropout)
         )
 
-    def forward(self, x, context = None):
+    def forward(self, x, context = None, mask = None):
         h = self.heads
 
         q = self.to_q(x)
@@ -80,7 +93,15 @@ class Attention(nn.Module):
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h = h), (q, k, v))
 
         sim = einsum('b i d, b j d -> b i j', q, k) * self.scale
+
+        if exists(mask):
+            max_neg_value = -torch.finfo(sim.dtype).max
+            mask = repeat(mask, 'b j -> (b h) () j', h = h)
+            sim.masked_fill_(~mask, max_neg_value)
+
+        # attention, what we cannot get enough of
         attn = sim.softmax(dim = -1)
+
         out = einsum('b i j, b j d -> b i d', attn, v)
         out = rearrange(out, '(b h) n d -> b n (h d)', h = h)
         return self.to_out(out)
@@ -93,7 +114,8 @@ class Perceiver(nn.Module):
         *,
         num_fourier_features,
         depth,
-        num_latents = 6,
+        input_channels = 3,
+        num_latents = 512,
         cross_dim = 512,
         latent_dim = 512,
         cross_heads = 1,
@@ -102,33 +124,46 @@ class Perceiver(nn.Module):
         latent_dim_head = 64,
         num_classes = 1000,
         attn_dropout = 0.,
-        ff_dropout = 0.
+        ff_dropout = 0.,
+        weight_tie_layers = False
     ):
         super().__init__()
 
         self.num_fourier_features = num_fourier_features
-        input_dim = (num_fourier_features * 2) + 1
+        input_dim = ((num_fourier_features * 2) + 1) * input_channels
+
         self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))
+        self.pos_emb = nn.Parameter(torch.randn(num_latents, latent_dim))
+
+        get_cross_attn = lambda: PreNorm(latent_dim, Attention(latent_dim, input_dim, dropout = attn_dropout), context_dim = input_dim)
+        get_cross_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim, dropout = ff_dropout))
+        get_latent_attn = lambda: PreNorm(latent_dim, Attention(latent_dim, dropout = attn_dropout))
+        get_latent_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim, dropout = ff_dropout))
+
+        if weight_tie_layers:
+            get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff = map(cache_fn, (get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff))
 
         self.layers = nn.ModuleList([])
         for _ in range(depth):
             self.layers.append(nn.ModuleList([
-                PreNorm(latent_dim, Attention(latent_dim, input_dim, dropout = attn_dropout), context_dim = input_dim),
-                PreNorm(latent_dim, FeedForward(latent_dim, dropout = ff_dropout)),
-                PreNorm(latent_dim, Attention(latent_dim, dropout = attn_dropout)),
-                PreNorm(latent_dim, FeedForward(latent_dim, dropout = ff_dropout))
+                get_cross_attn(),
+                get_cross_ff(),
+                get_latent_attn(),
+                get_latent_ff()
             ]))
 
         self.to_logits = nn.Linear(latent_dim, num_classes)
 
-    def forward(self, data):
+    def forward(self, data, mask = None):
         b = data.shape[0]
         data = fourier_encode(data, self.num_fourier_features)
+        data = rearrange(data, 'b n ... -> b n (...)')
 
-        x = repeat(self.latents, 'n d -> b n d', b = b)
+        x = self.latents + self.pos_emb
+        x = repeat(x, 'n d -> b n d', b = b)
 
         for cross_attn, cross_ff, latent_attn, latent_ff in self.layers:
-            x = cross_attn(x, context = data) + x
+            x = cross_attn(x, context = data, mask = mask) + x
             x = cross_ff(x) + x
             x = latent_attn(x) + x
             x = latent_ff(x) + x
